import argparse

# delegate to render_cli when called as a script
if __name__ == "__main__":
    from gamma.config.render_cli import render_cli

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers()
    render_cli(subparsers.add_parser("render", help="Render config values"))

    args = parser.parse_args()
    func = getattr(args, "func", None)
    if func:
        func(args)
    else:
        parser.print_help()



from collections.abc import MutableMapping, MutableSet
from operator import __eq__, __lt__
from typing import Generic, Iterable, List, Tuple, TypeVar

KT = TypeVar("KT")
VT = TypeVar("VT")


class PODict(MutableMapping, Generic[KT, VT]):
    """Dictionary that stores keys sorted (like a prio queue) that works
    for partial orders"""

    data: List[Tuple[KT, VT]]
    __unset = object()

    def __init__(self, data=None, lt=__lt__, eq=__eq__) -> None:
        self.data = []
        self.el_lt = lt
        self.el_eq = eq
        if data:
            # initialize
            items = getattr(data, "items", lambda: data)
            for k, v in items():
                self[k] = v

    def __setitem__(self, key: KT, value: VT) -> None:
        for i, (k, _) in enumerate(self.data):
            if self.el_eq(key, k):
                self.data.pop(i)
                self.data.insert(i, (key, value))
                return
            elif self.el_lt(key, k):
                self.data.insert(i, (key, value))
                return
        self.data.append((key, value))

    def __getitem__(self, key: KT) -> VT:
        for (k, v) in self.data:
            if self.el_eq(key, k):
                return v
        raise KeyError(key)

    def pop(self, key: KT, default=__unset) -> VT:
        try:
            item = self.popitem(key)
        except KeyError:
            if default is self.__unset:
                raise
            return default
        else:
            return item[1]

    def popitem(self, key: KT, default=__unset) -> Tuple[KT, VT]:
        for i, (k, v) in enumerate(self.data):
            if self.el_eq(key, k):
                self.data.pop(i)
                return k, v

        if default is not self.__unset:
            return default
        raise KeyError(key)

    def __iter__(self):
        return self.keys()

    def __len__(self) -> int:
        return self.data.__len__()

    def __delitem__(self, key: KT) -> None:
        for i, (k, _) in enumerate(self.data):
            if key == k:
                self.data.pop(i)
                return

    def keys(self) -> Iterable[KT]:
        return (x[0] for x in self.data)

    def items(self) -> Iterable[KT]:
        return iter(self.data)

    def values(self) -> Iterable[VT]:
        return (x[1] for x in self.data)

    def __contains__(self, o) -> bool:
        return o in self.keys()

    def clear(self) -> None:
        return self.data.clear()


class POSet(PODict[KT, None], MutableSet, Generic[KT]):
    def __init__(self, data, lt=__lt__, eq=__eq__) -> None:
        d = []
        for k in data:
            d.append((k, None))
        super().__init__(d, lt, eq)

    def add(self, value) -> None:
        self[value] = None

    def discard(self, value) -> None:
        del self[value]



import inspect
import types
from typing import Any, Generic, Tuple, TypeVar

T = TypeVar("T")


class ParametricMeta(type):
    def __getitem__(cls: T, key) -> T:
        return cls.of(key)

    def __new__(metacls, name, bases, namespace):
        @classmethod
        def of(cls, *values, name=None):
            if values not in cls._subclasses:
                values_s = ",".join([repr(x) for x in values])
                c_super = f"{cls.__name__}[{values_s}]"
                if name:
                    c_repr = "<instance of {name}>"
                    c_name = name
                    c_module = inspect.getmodule(inspect.currentframe().f_back).__name__
                else:
                    c_repr = f"{c_super}"
                    c_name = c_super
                    c_module = cls.__module__

                SubClass = types.new_class(c_name, (cls,))
                SubClass.__repr__ = lambda *_: c_repr
                SubClass.__module__ = c_module
                SubClass.__values__ = values
                SubClass.__tag__ = cls
                SubClass.__valuetype__ = True

                @classmethod
                def _also_of(_, *new_values):
                    cls._subclasses[new_values] = SubClass
                    return SubClass

                SubClass.also_of = _also_of
                cls._subclasses[values] = SubClass
            else:
                SubClass = cls._subclasses[values]
                if name and SubClass.__name__ != name:
                    raise NameError(
                        f"Parametric class with params {values} already instantiated "
                        f"with name {SubClass.__name__}"
                    )

            return SubClass

        @classmethod
        def any_of(cls, *values, name=None):
            first = cls.of(values[0], name=name)
            for alt in values[1:]:
                first.also_of(alt)

        Class = super().__new__(metacls, name, bases, namespace)
        Class._subclasses = dict()
        Class.of = of
        Class.any_of = any_of
        return Class


class parametric(Generic[T]):
    """Decorator to create new parametric base classes"""

    def __new__(deco, Class: T) -> T:
        Para = types.new_class(
            Class.__name__, (Class,), kwds={"metaclass": ParametricMeta}
        )
        module = inspect.getmodule(inspect.currentframe().f_back).__name__
        Para.__module__ = module
        return Para


class ValMeta(ParametricMeta):
    @property
    def values(cls) -> Tuple:
        """Return all values used to create the parametric value type"""
        return cls.__values__

    @property
    def value(cls) -> Any:
        vals = cls.values
        return vals and vals[0] or None


class Val(metaclass=ValMeta):
    """Generic parametric class with easy value accessors"""



from gamma.config.confignode import ConfigNode
from gamma.dispatch import dispatch
from ruamel.yaml.nodes import MappingNode, SequenceNode

from .render import render_node
from .tags import Map, Seq


def _prepare_ctx(**ctx):
    from gamma.config import get_config

    ctx.setdefault("recursive", True)
    ctx.setdefault("config", get_config())
    return ctx


@dispatch
def to_dict(node, **ctx):
    """Converts a node to a dictionary."""
    ctx = _prepare_ctx(**ctx)
    return render_node(node, **ctx)


@dispatch
def to_dict(node: ConfigNode, **ctx):
    """Converts a ConfigNode to a dictionary."""
    ctx.setdefault("config", node)
    ctx = _prepare_ctx(**ctx)
    return render_node(node, **ctx)


@dispatch
def to_dict(node: MappingNode, **ctx):
    """Render MappingNodes as dict regardless of tag value"""
    ctx = _prepare_ctx(**ctx)
    return render_node(node, Map(), **ctx)


@dispatch
def to_dict(node: SequenceNode, **ctx):
    """Render SequenceNodes as list regardless of tag value"""
    ctx = _prepare_ctx(**ctx)
    return render_node(node, Seq(), **ctx)



# flake8: noqa

__version__ = "0.5.0"

from ruamel.yaml.nodes import MappingNode, Node, ScalarNode, SequenceNode

from .builtin_tags import yaml
from .confignode import ConfigNode, RootConfig, config_context, push_entry, remove_entry
from .dump_dict import to_dict
from .dump_yaml import to_yaml
from .globalconfig import get_config
from .render import render_node
from .render_context import ContextVar, context_providers
from .tags import Tag



"""Module for discovering configuration files (entries)"""

import logging
import os
from pathlib import Path
from typing import Any, List, Optional, Tuple

from gamma.dispatch import Val, dispatch

from .confignode import ConfigNode
from .load import load_node

logger = logging.getLogger(__name__)

FindEnv = Val["env"]
FindLocal = Val["local"]
FindJupyter = Val["jupyter"]

CONFIG_FIND_ORDER = [FindEnv, FindLocal, FindJupyter]
CONFIG_ROOT_ENV = "GAMMA_CONFIG_ROOT"


@dispatch
def get_entries() -> List[Tuple[str, Any]]:
    """Discover the config root folder and get all entries"""
    config_root = get_config_root()
    load_dotenv(config_root)
    return get_entries(config_root)


@dispatch
def get_entries(folder: Path, *, meta_include_folders=True) -> List[Tuple[str, Any]]:
    """Get all entries in a given folder.

    Args:
        meta_include_folder: If `True`, try to load the `00-meta.yaml` file in the
            folder and follow `include_folder` entries.
    """
    paths = list(folder.glob("*.yaml")) + list(folder.glob("*.yml"))
    entries = [(p.name, p) for p in paths]
    if meta_include_folders:
        meta = load_meta(folder)
        for subfolder in meta["include_folders"]:
            if subfolder is None:
                continue
            subpath = folder / subfolder
            for k, e in get_entries(subpath, meta_include_folders=False):
                entries.append((k, e))

    return entries


@dispatch
def load_meta(config_root: Path) -> ConfigNode:
    """Load the `00-meta.yaml` file in a given folder"""
    meta_path = config_root / "00-meta.yaml"
    with meta_path.open("r") as fo:
        return ConfigNode(load_node(fo))


@dispatch
def get_config_root() -> Path:
    """Return the location for config root path.

    The mechanisms used to find the config root are set in the ``CONFIG_LOAD_ORDER``
    module variable.
    """

    for mech in CONFIG_FIND_ORDER:
        root = get_config_root(mech())
        if root is not None:
            return root
    raise Exception("Cannot locate a gamma config root in any expected location.")


@dispatch
def get_config_root(_: FindEnv) -> Optional[Path]:
    f"""Try the path set by {CONFIG_ROOT_ENV} as a root config folder"""
    if CONFIG_ROOT_ENV not in os.environ:
        return None

    root = Path(os.getenv(CONFIG_ROOT_ENV)).absolute()
    if not _has_meta(root):
        raise ValueError(
            f"Could not find {root}/00-meta.yaml file as pointed by "
            f"{CONFIG_ROOT_ENV} environment variable"
        )

    return root


@dispatch
def get_config_root(_: FindLocal) -> Optional[Path]:
    """Try the path `$PWD/config` as root config folder"""
    root = Path("config").absolute()
    if _has_meta(root):
        return root
    return None


@dispatch
def get_config_root(_: FindJupyter) -> Optional[Path]:
    """Try `<parent>/config` folders iteratively if we're in a Jupyter (IPython)
    environment"""

    if not _isnotebook():
        return None

    path = Path(".").absolute()
    while path != path.parent:
        candidate = path / "config"
        if _has_meta(candidate):
            return candidate
        path = path.parent

    return None


def _has_meta(root: Path):
    return (root / "00-meta.yaml").exists()


def _isnotebook():  # pragma: no cover
    try:
        shell = get_ipython().__class__.__name__  # type: ignore
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return True  # Terminal running IPython
        else:
            return True  # Other type of IPython kernel (?)
    except NameError:
        return False  # Probably standard Python interpreter


def load_dotenv(root: Path = None):
    """Load dotenv files located in:

    - `$PWD/config.local.env`
    - `{config_root}/../config.local.env`
    - `$PWD/config.env`
    - `{config_root}/../config.env`

    """

    import dotenv

    home = root.parent
    dotenv.load_dotenv("./config.local.env")
    dotenv.load_dotenv(f"{home}/config.local.env")
    dotenv.load_dotenv("./config.env")
    dotenv.load_dotenv(f"{home}/config.env")



"""Module declaring a cache utility for gamma.config"""

from typing import Any, Mapping


class Cache(Mapping):
    """
    A cache backed by a in-memory `dict`
    """

    def __init__(self) -> None:
        self.store = dict()

    def __getitem__(self, key) -> Any:
        return self.store.__getitem__(key)

    def __iter__(self):  # pragma: no cover
        return self.store.__iter__()

    def __setitem__(self, key, value):
        return self.store.__setitem__(key, value)

    def __len__(self) -> int:  # pragma: no cover
        return self.store.__len__()

    def clear(self):
        """Clear cache contents"""
        return self.store.clear()


cache = Cache()



"""Module to 'scaffold' append bootstrap config for gamma-io"""
import shutil
from pathlib import Path
from typing import NamedTuple

import pkg_resources

import colorama
from colorama import Fore, Style
from gamma.dispatch import dispatch

ENTRYPOINT_GROUP = "gamma.config.scaffold"


class GammaConfigScaffold(NamedTuple):
    name = "gamma-config"


@dispatch
def get_source(module: GammaConfigScaffold):
    return Path(__file__).parent / "sample"


@dispatch
def get_files(module, src):
    return sorted(src.glob("**/*"))


def scaffold(target, force):
    """Initialize the config folder with samples.

    The set of files to be deployed on scaffolding can be extended via the Python
    "entrypoint" plugin system.

    First you need to add your the directives in `setup.cfg` pointing to the
    `gamma.config.scaffold` group. Example:

    ```ini
    [options.entry_points]
    gamma.config.scaffold =
        gamma-io = gamma.io.scaffold:setup
    ```

    In the example above, `setup` function under `gamma.io.scaffold` module should
    return an instance of any class that will be used for dispatching the `get_files`
    and `get_source` functions.
    """

    colorama.init()  # for fellow windows users

    if not target:
        target = Path(".").absolute()

    print(f"TARGET: {target}")
    target = Path(target)
    confdir: Path = target / "config"
    confdir.mkdir(exist_ok=True)

    # load plugins from ENTRYPOINT_GROUP
    modules = [GammaConfigScaffold()]
    for entry in pkg_resources.iter_entry_points(ENTRYPOINT_GROUP):
        plugin = entry.load()
        modules.append(plugin())

    # find source under sample folder
    for mod in modules:
        src = get_source(mod)
        for srcfile in get_files(mod, src):
            # Check if target file already exists
            relfile = srcfile.relative_to(src)
            dstfile: Path = confdir / relfile
            if srcfile.is_dir():
                continue
            elif not force and dstfile.exists():
                print(
                    Fore.YELLOW + f"[{mod.name}] File '{relfile}' already exists "
                    "and --force was not set",
                    Style.RESET_ALL,
                )
            else:
                dstfile.parent.mkdir(parents=True, exist_ok=True)
                shutil.copyfile(srcfile, dstfile)
                _file = Fore.CYAN + str(relfile) + Style.RESET_ALL
                print(
                    f"[{mod.name}] copied file {_file} to",
                    str(confdir.relative_to(target)),
                )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Initialize config folder with samples"
    )
    parser.add_argument(
        "-t",
        "--target",
        help="The target folder. Default to current folder.",
        default=None,
    )
    parser.add_argument(
        "--force",
        help="If set, do not check for existing files.",
        action="store_true",
        default=False,
    )

    args = parser.parse_args()
    from gamma.config.scaffold import scaffold as _run

    _run(args.target, args.force)



"""Module dealing with the lifecycle of the global `RootConfig` object"""

import multiprocessing
import threading
from typing import Optional, Tuple

from .cache import cache
from .confignode import RootConfig, push_entry
from .findconfig import get_entries
from .load import load_node


class _GlobalStore:
    """Implements a global store holder object.

    This store operates as a restricted version of a thread-local, where you can only
    ovewrite the value in the same thread/processes where it was first set.
    """

    def __init__(self) -> None:
        self.root = None
        self.load_key = None
        self.load_thread_name = None
        self.load_pid = None

    def set(self, root, force=False) -> None:
        """Set the new root config.

        To avoid unwanted suprises, by defaul we allow updating the global root config
        only by the same thread that created it.

        Args:
            root: the RootConfig to store
            force: if True, do not check for thread safety
        """

        key = self._get_load_key()
        if not force:
            self.check_can_modify(key=key)

        self.root = root
        self.load_key = key
        self.load_thread_name = threading.current_thread().name

    def get(self) -> RootConfig:
        return self.root

    def empty(self) -> bool:
        return self.root is None

    def reset(self, force=False) -> None:
        if not force:
            self.check_can_modify()
        self.__init__()

    def _get_load_key(self) -> Tuple[int, int]:
        thread = threading.current_thread()
        pid = multiprocessing.current_process().pid
        key = (thread.ident, pid)
        return key

    def check_can_modify(self, key=None) -> None:
        """Raises an Exception modifying the config would break the same thread rule.

        Args:
            key: the key to check, fetches one if `None`
        """
        key = key or self._get_load_key()
        if self.root is None or self.load_key == key:
            return

        thread = threading.current_thread()
        raise Exception(
            f"Current thread {thread.name}(tid={key[0]}; pid={key[1]}) tried to update "
            f"the global config that was originally loaded from thread "
            f"{self.load_thread_name}(tid={self.load_key[0]}; pid={self.load_key[1]})"
        )


_global_store = _GlobalStore()


def get_config(initialize: bool = True) -> Optional[RootConfig]:
    """Get the global config root object, loading if needed and `initialize` is `True`.

    This global object is cached and safe to call multiple times, from multiple
    threads.
    """
    if _global_store.empty() and not initialize:
        return None
    elif _global_store.empty() and initialize:
        entries = sorted(get_entries())
        root = RootConfig()
        for entry_key, entry in entries:
            node = load_node(entry)
            if node:
                push_entry(root, entry_key, node)

        _global_store.set(root)
    return _global_store.get()


def reset_config(force: bool = False) -> None:
    """Clear the global store and cache.

    This is not meant to be used extensively by applications, but can be useful for
    writing tests.

    Args:
        force: if True, will reset the global store regardless of the current
            thread/process.
    """

    _global_store.reset(force=force)
    cache.clear()


def set_config(cfg: RootConfig) -> None:
    """Forces a root config.

    Call `reset_config`.

    NOTE: This is not meant to be used regularly by applications, but can be useful
    when writing tests.
    """
    reset_config()
    _global_store.set(cfg)


def check_can_modify(cfg: RootConfig) -> None:
    """Check if this root config object can be modified.

    As specified, the **global** root config object can only be modified by the thread
    that first created it.
    """

    if _global_store.root is cfg:
        _global_store.check_can_modify()



"""Module that implements renderers for builtin-tags"""

import importlib
import os
import threading
from typing import Any, Union

from gamma.config.confignode import ConfigNode
from gamma.config.dump_dict import to_dict
from gamma.dispatch import dispatch
from jinja2.runtime import StrictUndefined
from ruamel.yaml import YAML
from ruamel.yaml.nodes import MappingNode, Node, ScalarNode, SequenceNode

from .findconfig import get_config_root
from .render import render_node
from .render_context import get_render_context
from .tags import Tag, TagException

UNDEFINED = "~~UNDEFINED~~"
yaml = YAML(typ="safe")


RefTag = Tag["!ref"]
EnvTag = Tag["!env"]
EnvSecretTag = Tag["!env_secret"]
ExprTag = Tag["!expr"]
J2Tag = Tag["!j2"]
J2SecretTag = Tag["!j2_secret"]
PyTag = Tag["!py"]
ObjTag = Tag["!obj"]
PathTag = Tag["!path"]

j2_cache = threading.local()


@dispatch
def render_node(node: Node, tag: EnvTag, **ctx) -> str:
    """[!env] Maps the value to an environment variable of the same name.

    You can provide a default using the ``|`` (pipe) character after the variable
    name.

    Examples:

        my_var: !env MYVAR|my_default
    """
    name = node.value
    name, default = _split_default(name)
    env_val = os.getenv(name, default)
    if env_val == UNDEFINED:
        raise TagException(
            f"Env variable '{name}' not found when resolving node and no default set"
        )

    return env_val


# process: !env_secret
@dispatch
def render_node(node: Node, tag: EnvSecretTag, *, dump=False, **ctx) -> str:
    """[!env_secret] Similar to !env, but never returns the value when dumping."""
    if dump:
        return node
    return render_node(node, EnvTag(), dump=dump, **ctx)


# process: !expr
@dispatch
def render_node(node: Node, tag: ExprTag, **ctx) -> Any:
    """[!expr] Uses ``eval()`` to render arbitrary Python expressions.

    By default, we add the root configuration as `c` variable.

    See ``gamma.config.render_context.context_providers`` documentation to add your
    own variables to the context.
    """

    _locals = {}
    _globals = get_render_context(node=node, tag=tag, **ctx)
    return eval(node.value, _globals, _locals)


@dispatch
def render_node(node: Node, tag: J2Tag, *, config=None, key=None, **ctx) -> Any:
    """[!j2] Treats the value a Jinj2 Template

    See ``gamma.config.render_context.context_providers`` documentation to add your
    own variables to the context.

    In practice, in the snippet bellow, ``foo1`` and ``foo2`` are equivalent

        myvar: 100
        foo1: !expr f"This is a number = {c.myvar}"
        foo2: !j2 This is a number = {c.myvar}

    You can customize the Jinja2 environment by providing a reference to a Python
    function in the `j2_env` key in `00-meta.yaml`. Example

        j2_env: my_app.my_module:my_func

    Notes:
        * Jinja2 is not installed by default, you should install it manually.
    """
    try:
        import jinja2
        import jinja2.exceptions
    except ModuleNotFoundError:  # pragma: no cover
        raise Exception(
            "Could not find Jinja2 installed. You must manually install it with "
            "`pip install jinja2` if you want to use the !j2 tag"
        )

    render_ctx = get_render_context(node=node, tag=tag, config=config, key=key, **ctx)

    if not hasattr(j2_cache, "env"):
        root = config and config._root
        env_factory = root.get("j2_env")
        if env_factory:
            module_name, func_name = env_factory.split(":", 1)
            mod = importlib.import_module(module_name)
            func = getattr(mod, func_name)
            j2_cache.env = func()
        else:
            j2_cache.env = jinja2.Environment(undefined=StrictUndefined)

    try:
        res = j2_cache.env.from_string(node.value).render(**render_ctx)
    except jinja2.exceptions.UndefinedError as ex:
        msg = f'Error rendering key `{key.value}: "{node.value}"` -> {ex.message}'
        raise ValueError(msg)

    return res


# process: !j2_secret
@dispatch
def render_node(node: Node, tag: J2SecretTag, *, dump=False, **ctx) -> Any:
    """[!j2_secret] Similar to !j2, but never returns the value when dumping."""

    if dump:
        return node
    return render_node(node, J2Tag(), dump=dump, **ctx)


@dispatch
def render_node(node: Node, tag: RefTag, *, config=None, recursive=False, **ctx) -> Any:
    """[!ref] References other entries in the config object.

    Navigate the object using the dot notation. Complex named keys can be accessed
    using quotes.
    """

    import functools
    import operator
    import shlex

    lex = shlex.shlex(instream=node.value, posix=True)
    lex.whitespace = "."
    tokens = []
    token = lex.get_token()
    while token:
        tokens.append(token)
        token = lex.get_token()

    parent = functools.reduce(operator.getitem, tokens[:-1], config._root)
    val = parent[tokens[-1]]

    if recursive:
        while isinstance(val, (ConfigNode, Node)):
            # chain dump if needed
            val = render_node(val, config=config, recursive=recursive, **ctx)

    return val


# process: !py
@dispatch
def render_node(node: ScalarNode, tag: PyTag, *, path=None, **ctx) -> Any:
    """[!py] Pass the node value to a Python callable.

    This tag should be used as a URI-style tag on the form `!py:<module>:<callable>`

    The scalar node value is first implicitly resolved to a Python object using
    `yaml.load`. For instance:

    ```yaml
    foo: !py:myapp.mymodule:myfunc 100
    bar: !py:myapp.mymodule:myfunc "100"
    zig: !py:myapp.mymodule:myfunc a value
    ```

    Will call the function `myfunc` in `myapp.mymodule` module with the arguments:
    - `type(value) == int; value == 100` for `foo`
    - `type(value) == str; value == "100"` for `bar`
    - `type(value) == str; value == "a value"` for `zig`
    """

    func = _py_tag_get_func("py", path)
    val = yaml.load(node.value)
    return func(val)


# process: !py
@dispatch
def render_node(
    node: Union[SequenceNode, MappingNode], tag: PyTag, *, path=None, **ctx
) -> Any:
    """[!py] Pass the node value to a Python callable.

    This tag should be used as a URI-style tag on the form `!py:<module>:<callable>`

    The `map` or `seq` node value is first converted to a Python `dict`/`list`
    recursively using the `to_dict` method.
    """

    func = _py_tag_get_func("py", path)
    val = to_dict(node, **ctx)
    return func(val)


def _py_tag_get_func(tag, path, default_module=None):
    """Get the callable for a given tag `path` argument

    Used by `!py`, `!obj` tags meta config.

    Args:
        tag: either "obj" or "py"
        path: the tag "path" part in the format `!<tag>:<module>:<callable>`
    """

    import importlib

    if not path:
        raise ValueError(f"The !{tag} tag is missing the 'path' part after the ':'")

    if ":" in path:
        module_name, func_name = path.split(":", 1)
    elif tag == "obj" and default_module:
        module_name = default_module
        func_name = path
    else:
        raise ValueError(f"The !{tag} tag has an invalid reference: {path}")

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as ex:
        raise ValueError(f"Module part from '!{tag}:{path}' not found") from ex

    func = getattr(module, func_name, None)
    if not func:
        raise ValueError(f"Name '{func_name}' not found in module {module}")

    return func


# process: !obj
@dispatch
def render_node(
    node: MappingNode, tag: ObjTag, *, path=None, config=None, **ctx
) -> Any:
    """[!obj] Create a Python object by passing the mapping value as nested dicts to
    the object constructor.

    This tag should be used as a URI-style tag on the form `!obj:<module>:<callable>`
    It behaves like `!py`, but only applies to `map` nodes, and
    automatically unpack the mapping.

    You may omit the `<module>` part of the path if you define an `obj_default_module`
    scalar entry at the root of the config.

    Examples:

    ```yaml
    foo: !obj:myapp.mymodule:MyClass
        a: 100
        b: a value

    # the above `foo` is equivalent to `bar` below

    obj_default_module: myapp.mymodule

    bar: !obj:MyClass
        a: 100
        b: a value
    ```
    """

    root = config and config._root
    default_module = (root and root.get("obj_default_module")) or None
    func = _py_tag_get_func("obj", path, default_module=default_module)
    val = to_dict(node)
    return func(**val)


# process: !path
@dispatch
def render_node(node: Node, tag: PathTag, **ctx) -> str:
    """[!path] Construct an absolute file path by joining a path
    fragment to the known path of the *parent* of config root directory

    Examples:
        # should point to `<config-root>/../data/hello_world.csv`
        my_var: !path data/hello_world.csv
    """
    path_fragment = node.value
    base = get_config_root().parent
    return str(base.joinpath(path_fragment).absolute())


###
# Utilities
###


def _split_default(value):
    """Get a value optional default, using the content after the last "|"
    (pipe) as convention.

    The default will be parsed as a YAML value.
    """

    default = UNDEFINED
    parsed = value
    if "|" in value:
        parsed, default = value.rsplit("|", 1)
        default = yaml.load(default)
    return parsed, default



"""Module for loading content as `ruamel.yaml` `Node` instances"""
from io import IOBase, StringIO
from pathlib import Path
from typing import Dict

from gamma.dispatch import dispatch, parametric
from ruamel.yaml.nodes import Node


@parametric
class ContentType:
    pass


YAMLContent = ContentType["text/yaml"]


@dispatch
def load_node(val: str, content_type=None) -> Node:
    """Load a string as `Node`, defaults to YAML content"""
    stream = StringIO(val)
    stream.seek(0)
    return load_node(stream, content_type)


@dispatch
def load_node(stream: IOBase, content_type=None) -> Node:
    """Load a stream as `Node`, defaults to YAML content"""
    content_type = content_type or YAMLContent()
    return load_node(stream, content_type)


@dispatch
def load_node(stream: IOBase, _: YAMLContent) -> Node:
    """Load a YAML stream"""
    from ruamel.yaml import YAML

    yaml = YAML()
    const, _ = yaml.get_constructor_parser(stream)
    node = const.composer.get_single_node()
    return node


@dispatch
def load_node(entry: Path, content_type=None) -> Node:
    """Load path's content, defaults to YAML content"""
    try:
        content: str = entry.read_text().strip()
        if not content:
            return load_node({})
        return load_node(content, content_type)
    except Exception as ex:
        raise ValueError(f"Error loading from file '{entry}'") from ex


@dispatch
def load_node(val: Dict, content_type=None) -> Node:
    """Load dict as node"""
    from ruamel.yaml import YAML

    if not val:
        val = dict()
    yaml = YAML()
    stream = StringIO()
    yaml.dump(val, stream)
    stream.seek(0)
    content = stream.read()
    return load_node(content, YAMLContent())



"""Module implementing convenience methods for dealing with `ruamel.yaml` `Node`s"""
from typing import Any, Hashable, Iterable, Optional, Tuple

from gamma.dispatch import dispatch
from ruamel.yaml.nodes import MappingNode, Node, ScalarNode, SequenceNode

Entry = Tuple[Node, Optional[Node]]


@dispatch
def get_keys(node: MappingNode) -> Iterable[Node]:
    """Get all keys on a `map` node"""
    for item_key, _ in node.value:
        yield item_key


@dispatch
def get_item(node: MappingNode, key, *, default=...) -> Node:
    """Get a single child node item from `map` node"""
    for item_key, item_value in node.value:
        if is_equal(key, item_key):
            return item_value

    if default is not Ellipsis:
        return default

    raise KeyError(key)  # pragma: no cover


@dispatch
def get_entry(node: MappingNode, key, *, default=...) -> Entry:
    """Get a (key, value) entry from a `map` node.

    Args:
        default: return this value instead of `KeyError` if not found.

    Raise:
        `KeyError` if key not found and `default` not provided
    """
    for item_key, item_value in node.value:
        if is_equal(key, item_key):
            return item_key, item_value

    if default is not Ellipsis:
        return as_node(key), default

    raise KeyError(key)


@dispatch
def is_equal(a: ScalarNode, b) -> bool:
    """Check if `a` is equal to `b`"""
    return is_equal(a, as_node(b))


@dispatch
def is_equal(a, b: ScalarNode) -> bool:
    """Check if `a` is equal to `b`"""
    return is_equal(as_node(a), b)


@dispatch
def is_equal(a: ScalarNode, b: ScalarNode):
    return a.tag == b.tag and a.value == b.value


@dispatch
def as_node(a: Node) -> Node:
    return a


@dispatch
def as_node(a) -> Node:
    """Return the `Node` representation of a given object.

    This method handle basic scalar types.
    """

    # handle base types
    if isinstance(a, str):
        return ScalarNode("tag:yaml.org,2002:str", value=a)
    elif isinstance(a, bool):
        return ScalarNode("tag:yaml.org,2002:bool", value=str(a))
    elif isinstance(a, int):
        return ScalarNode("tag:yaml.org,2002:int", value=str(a))
    elif isinstance(a, float):
        return ScalarNode("tag:yaml.org,2002:float", value=str(a))
    elif a is None:
        return ScalarNode("tag:yaml.org,2002:null", value="null")
    else:
        raise Exception(f"Can't handle type {type(a)} for value {a}")


@dispatch
def get_id(a) -> Hashable:
    # fallback
    return a


@dispatch
def get_id(a: ScalarNode) -> Hashable:
    """Return an object that allows to ScalarNodes to be hashed and compared"""
    return (a.tag, a.value)


@dispatch
def is_in(node: Node, container: Iterable[Any]) -> bool:
    """Return true if `node` is in `container`"""
    for item in container:
        if is_equal(item, node):
            return True
    return False


@dispatch
def get_entries(node: MappingNode) -> Iterable[Entry]:
    """Return all entries (key, value) in this `map` node"""
    for item_key, item_value in node.value:
        yield item_key, item_value


@dispatch
def get_values(node: SequenceNode) -> Iterable[Node]:
    """Return all values in this `seq` node"""
    for item_value in node.value:
        yield item_value


@dispatch
def get_values(node: MappingNode) -> Iterable[Node]:
    """Return all values in this `map` node"""
    for _, item_value in node.value:
        yield item_value


@dispatch
def union_nodes(first: Iterable, second: Iterable) -> Iterable:
    """Union two sets of nodes.

    By default we keep the ones in `first` if equals."""

    out = dict()

    for a in first:
        out[get_id(a)] = a

    for b in second:
        if b not in out:
            out[get_id(b)] = b

    return list(out.values())



import copy
from io import StringIO

from gamma.dispatch import dispatch
from ruamel.yaml import YAML
from ruamel.yaml.nodes import MappingNode, ScalarNode, SequenceNode

from .confignode import ConfigNode, RootConfig
from .merge import merge_nodes
from .rawnodes import as_node
from .render import render_node
from .tags import Tag


def yaml_serialize(node):
    with StringIO() as stream:
        yaml = YAML(typ="rt")
        yaml.serialize(node, stream)
        stream.seek(0)
        return stream.read()


@dispatch
def to_yaml(cfg: RootConfig, resolve_tags: bool):
    """Dump a config/node to the YAML representation.

    Args:
        resolve_tags: if True, will render tags, otherwise, dump tags unrendered.
    """

    nodes = list(cfg._root_nodes.values())
    _, node = merge_nodes(nodes)
    if resolve_tags:
        node = dump_node(node, config=cfg)
    return yaml_serialize(node)


@dispatch
def to_yaml(cfg: ConfigNode, resolve_tags: bool):
    node = cfg._node
    if resolve_tags:
        node = dump_node(node, config=cfg)
    return yaml_serialize(node)


@dispatch
def to_yaml(cfg: ConfigNode):
    """Render a ConfigNode assuming default of `resolve_tags` = True"""
    return to_yaml(cfg, True, config=cfg)


@dispatch
def dump_node(node: ScalarNode, *, config=None):
    """Dump a `scalar` node as raw YAML node"""
    tag: str = node.tag
    if not tag.startswith("tag:yaml.org,2002:"):
        tagobj = Tag[tag]()
        val = render_node(node, tagobj, dump=True, config=config)
        return as_node(val)
    return node


@dispatch
def dump_node(node: SequenceNode, *, config=None):
    """Dump a `seq` node as raw YAML node"""

    tag: str = node.tag

    # render if sequence itself is tagged
    if not tag.startswith("tag:yaml.org,2002:"):
        tagobj = Tag[tag]()
        val = render_node(node, tagobj, dump=True, config=config)
        return as_node(val)

    # resolve recursively for each value
    newvalue = [dump_node(v, config=config) for v in node.value]
    node = copy.copy(node)
    node.value = newvalue
    return node


@dispatch
def dump_node(node: MappingNode, *, config=None):
    """Dump a `scalar` node as raw YAML node"""
    tag: str = node.tag

    # render if sequence itself is tagged
    if not tag.startswith("tag:yaml.org,2002:"):
        tagobj = Tag[tag]()
        val = render_node(node, tagobj, dump=True, config=config)
        return as_node(val)

    # resolve recursively for each entry
    newvalue = [(k, dump_node(v, config=config)) for (k, v) in node.value]
    node = copy.copy(node)
    node.value = newvalue
    return node



import json
from argparse import ArgumentParser, RawTextHelpFormatter

from gamma.config import to_dict
from gamma.config.load import load_node


def render_cli(parser: ArgumentParser):
    descr = """Render VALUE on the stdout.

The VALUE may incorporate a custom tag (eg. `!j2 '{{ c.foo }}'`). If not provided,
we assume the value is a prefixed with `!ref`.

Examples:

    # render the value of "foo.bar", equivalent to "!ref foo.bar"
    python gamma.config render foo.bar

    # render a custom Jinja2 templated value
    python gamma.config render "!j2 /path/to/{{ c.foo.bar }}"

"""
    parser.description = descr
    parser.formatter_class = RawTextHelpFormatter

    parser.add_argument("value")
    parser.add_argument(
        "-o",
        "--output",
        choices=["raw", "json"],
        default="json",
        help="Output format, defaults to 'json'",
    )

    def _func(args):
        print(_do_render(args.value, args.output))

    parser.set_defaults(func=_func)
    return parser


def _do_render(value: str, output="json"):
    """Render a value and return as string.

    Args:
        output: either `raw` or `json`. Note `raw` only makes sense for scalars
    """

    if not value.startswith("!"):
        value = "!ref " + value.strip()

    node = load_node(value)
    data = to_dict(node)
    if output == "raw":
        return data
    elif output == "json":
        return json.dumps(data, indent=2)
    else:
        raise ValueError(f"Invalid output format: {output}")


if __name__ == "__main__":
    render_cli()



"""Definition of base Tag class and standard YAML derived tag types"""
from gamma.dispatch import parametric


@parametric
class Tag:
    @property
    def name(self):
        return str(self.__values__[0])


Map = Tag.of("tag:yaml.org,2002:map")
Seq = Tag.of("tag:yaml.org,2002:seq")
Str = Tag.of("tag:yaml.org,2002:str")
Int = Tag.of("tag:yaml.org,2002:int")
Float = Tag.of("tag:yaml.org,2002:float")
Bool = Tag.of("tag:yaml.org,2002:bool")
Timestamp = Tag.of("tag:yaml.org,2002:timestamp")
Null = Tag.of("tag:yaml.org,2002:null")


class TagException(Exception):
    pass



"""Module handling rendering context variables (eg. for !expr and !j2)"""
from functools import partial
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Union

from gamma.config.confignode import ConfigNode

from .cache import cache


class ContextVar(NamedTuple):
    """A specification for a context variable used for rendering dynamic values.

    Either if ``function`` is specified, we'll call the function to get the actual
    value.
    """

    name: str
    """The name of the variable"""

    value: Any = None
    """The value of the variable"""

    function: Optional[Callable[..., Any]] = None
    """Function to call to resolve the variable"""

    cacheable: bool = False
    """If True, will cache the function result, otherwise will call on each render."""


def get_render_context(**kwargs) -> Dict[str, Any]:
    """Return the render context by calling each function in ``context_provider``.

    A context provider must be a function with the signature:
        `(**kwargs) -> List[ContextVar]`
    or simply a list of `ContextVar` objects

    The provided `**kwargs` are the same available in the `render_node` function
    """
    out = {}
    for provider in context_providers:
        var: ContextVar
        vars: List[ContextVar] = provider(**kwargs) if callable(provider) else provider
        for var in vars:
            if var.cacheable:
                cache_key = f"render_context/{var.name}"
                try:
                    out[var.name] = cache[cache_key]
                    continue
                except KeyError:
                    pass

            if var.function is not None:
                val = var.function()
            else:
                val = var.value

            if var.cacheable:
                cache[cache_key] = val

            out[var.name] = val

    return out


###
# Built-in providers
###


def base_provider(**kwargs):
    """Defaults to context

    - `env` -> os.environ
    - `c` -> the global RootConfig
    """
    import os
    from .globalconfig import get_config

    return [
        ContextVar("env", os.environ),
        ContextVar("c", function=partial(get_config, False)),
    ]


def underscore_context_provider(*, config: ConfigNode = None, **kwargs):
    """Look in parent config nodes and add all entries under the `_context` key"""

    vars = []
    if not config:
        return vars

    while config._parent is not None:
        config = config._parent
        ctx: dict = config.get("_context", {})
        for key, value in ctx.items():
            vars.append(ContextVar(key, value))

    return vars


context_providers: List[Union[Callable, List[ContextVar]]] = [
    base_provider,
    underscore_context_provider,
]
"""Provides a list of context providers. See :func:get_render_context() for details
"""



import functools
import inspect
import typing
import warnings
from collections import defaultdict
from typing import Callable, Dict, Iterable, List, Set, Tuple

from .typesystem import Sig, SigDict, is_more_specific, issubtype, signatures_from


class DispatchError(Exception):
    pass


class OverwriteWarning(UserWarning):
    pass


def methods_matching(call, table) -> List:
    """Given a method table, return the methods matching the call signature.

    Args:
        table: an iterable of methods ordered by `is_more_specific`
        call: the call signature as a `Sig` object or `Tuple[...]` type
    """
    matches = [x for x in table if issubtype(call, x)]
    methods = []
    while matches:
        match = matches[0]
        methods.append(match)
        matches = [x for x in matches[1:] if not is_more_specific(match, x)]
    return methods


def get_type(val):
    if isinstance(val, type) and getattr(val, "__valuetype__", False):
        return val
    elif isinstance(val, type):
        return typing.Type[val]
    else:
        return type(val)


_namespaces = dict()


class dispatch:
    """Function wrapper to dispatch methods.

    Args:
        namespace: If set, use a shared namespace with `namespace` as key, otherwise
            try to find matching functions in `locals`.
        specialize: If set, will throw an error if this function is not a specialization
            of an already existing function.
        overwrite: When `True`, won't issue a warning about overwriting a method.
    """

    #: Pending methods register due to forward references
    pending: Set

    #: The methods table for this function
    methods: SigDict[Callable]

    #: Cache from call signature to actual function
    cache: Dict[Tuple, Callable]

    #: Callable to get types from function arguments
    get_type: Callable

    #: function name
    name: str

    #: set of reserved argument names
    arg_names: Dict[str, List[Sig]]

    def __new__(cls, *args, namespace=None, specialize=False, overwrite=False):
        if isinstance(namespace, str):
            namespace: dict = _namespaces.setdefault(namespace, {})
            add_to_namespace = True
        else:
            namespace = inspect.currentframe().f_back.f_locals
            add_to_namespace = False

        def wrapped(func):
            # check if function with the same name exists in scope
            existing = namespace.get(func.__name__)
            if existing and isinstance(existing, dispatch):
                existing.register(func, overwrite=overwrite)
                return existing

            if specialize:
                raise DispatchError(
                    f"This method has `specialize=True` but no existing "
                    f"function named `{func.__name__}` exists."
                )

            # create a new dispatch table wrapper
            self: dispatch = functools.update_wrapper(object.__new__(cls), func)
            self.pending = set()
            self.get_type = get_type
            self.methods = SigDict()
            self.cache = dict()
            self.name = func.__name__
            self.arg_names = defaultdict(list)
            self.register(func, overwrite=overwrite)

            _dispatch_by_name[self.name] = self

            if add_to_namespace:
                namespace[func.__name__] = self

            return self

        if args and callable(args[0]):
            return wrapped(args[0])

        return wrapped

    def register(
        self, func: Callable, *, overwrite=False, allow_pending=True
    ) -> Callable:
        """Register a new method to this function's dispatch table.

        Args:
            func: the method to register
            overwrite: if False, will warn if the registration will overwrite and
                existing registration.
            allow_pending: if True, won't error on forward references
        """
        self.clear()
        try:
            for sig in signatures_from(func):
                self._register_single(sig, func, overwrite)
                for name in sig.arg_names:
                    lst = self.arg_names[name]
                    try:
                        lst.remove(sig)
                    except ValueError:
                        pass
                    lst.append(sig)
            return self
        except NameError:
            if allow_pending:
                self.pending.add((func, overwrite))
            else:
                raise

    def _register_single(self, sig, func, overwrite):
        old, _ = self.methods.popitem(sig, (None, None))
        if old is not None and not overwrite:
            self._warn_overwrite(sig, old)
        self.methods[sig] = func

    def _warn_overwrite(self, new: Sig, old: Sig):
        msg = (
            f"Method for function '{self.name}' defined\n"
            f"    at {old.func_site}\n"
            f"will be overwritten by new method defined\n"
            f"    at {new.func_site}\n"
            f"If using kw-or-positional vs kw-only arguments correctly."
        )
        warnings.warn(msg, OverwriteWarning)

    def clear(self):
        """Empty the cache."""
        self.cache.clear()

    def __setitem__(self, key, func: Callable):
        """Manually map a call signature to a callable"""

        self.clear()
        if not key:
            return
        key = key if isinstance(key, tuple) else (key,)
        sig = Sig(*key)
        self._register_single(sig, func, False)

    def __delitem__(self, types: Tuple):
        """Remove a method registration"""
        self.clear()
        sig = Sig(*types)
        self.methods.pop(sig)

    def __getitem__(self, key: Tuple):
        return self.find_method(key)

    def find_method(self, key: Tuple) -> Callable:
        """Find and cache the next applicable method of given types.

        Args:
            key: A call args types tuple.
        """

        # ensure tuple of types
        if getattr(key, "__origin__", None) is tuple:
            key = key.__args__
        elif not isinstance(key, tuple):
            if isinstance(key, Iterable):
                key = tuple(*key)
            else:
                key = tuple([key])

        self.resolve_pending()
        cached = self.cache.get(key)
        if cached:
            return cached

        call_sig = Sig(*key)

        match = methods_matching(call_sig, self.methods)
        if not match:
            self._except_no_method_found(key)

        if len(match) > 1:
            self._except_no_meet(key, match)

        method = self.methods[match[0]]

        # cache and return
        self.cache[key] = method
        return method

    def _except_no_meet(self, key, match: List[Sig]):
        names = ", ".join(":" + t.__name__ for t in key)
        msg = f"Method call ({names}) is ambiguous. Candidates:"
        for m in match:
            msg += f"\n    {m} at {m.func_site}"
        raise DispatchError(msg)

    def _except_no_method_found(self, key):
        names = ", ".join(":" + getattr(t, "__name__", repr(t)) for t in key)
        msg = f"{self}: no method found for call ({names})"
        msg = self._msg_multiple_definitions_same_name(msg, key)
        raise DispatchError(msg)

    def _msg_multiple_definitions_same_name(self, msg, key):
        other = []
        for name, _dispatch in _dispatch_by_name.items():
            if name == self.name and _dispatch is not self:
                other.append(_dispatch.find_method(key))

        if other:
            msg += (
                "\n\nNote: found maching method(s) with the same name "
                "but in another module. This may be an error due to a missing import."
            )
            for func in other:
                co = func.__code__
                msg += f"\n\n    -> {co.co_filename}:{co.co_firstlineno}"

            msg += "\n"
        return msg

    def __call__(self, *args, **kwargs):
        """Resolve and dispatch to best method."""
        try:
            if kwargs:
                self._check_invalid_kwargs(kwargs)
            key = tuple(self.get_type(a) for a in args)
            func = self.cache.get(key)
            if not func:
                func = self.find_method(key)

            try:
                return func(*args, **kwargs)
            except TypeError as ex:
                if "keyword argument" not in ex.args[0]:
                    raise
                file = func.__code__.co_filename
                line = func.__code__.co_firstlineno
                msg = f"{ex.args[0]}\n    in function {file}:{line}"
                raise TypeError(msg) from ex

        except DispatchError as ex:
            file = inspect.currentframe().f_back.f_code.co_filename
            line = inspect.currentframe().f_back.f_lineno
            msg = f"{ex.args[0]}\n    called from {file}:{line}"
            raise DispatchError(msg) from ex

    def _check_invalid_kwargs(self, kwargs):
        """Check if we're passing 'reserved' arg names in kwargs.

        Python allows passing positional args as kwargs. This effectively forbid this,
        ie. all keyword call arguments are deemed to be kwargs, and this is required to
        avoid weird behavior.

        Example:

            @dispatch
            def foo(a: int, *, b):
                return "ok"

            assert foo(1, b=1) == "ok"
            assert foo(1, b="foo") == "ok"
            assert foo(a=1, b=1) == "ok"   # <- this is not allowed
        """

        isec = set(kwargs).intersection(self.arg_names)
        if isec:
            msg = (
                f"Reserved argument names were used as keywords in "
                f"method call: {isec}"
            )
            for arg in isec:
                msg += f"\n    '{arg}' in signatures:"
                for sig in self.arg_names[arg]:
                    msg += f"\n        '{sig.dump()}'"

            raise DispatchError(msg)

    def resolve_pending(self):
        """Evaluate any pending forward references.

        This can be called explicitly when using forward references,
        otherwise cache misses will evaluate.
        """
        while self.pending:
            func, overwrite = self.pending.pop()
            self.register(func, overwrite=overwrite, allow_pending=False)

    def dump(self) -> str:  # pragma: no cover
        """Pretty-print debug information about this function"""
        msg = repr(self) + " with signatures:"
        for sig in self.methods:
            msg += "\n    " + sig.dump()
        return msg

    def __repr__(self) -> str:
        n = len(self.methods)
        p = ""
        if n == 1:
            p = "(s)"
        return f"<function '{self.__name__}' with {n} method{p}>"


_dispatch_by_name: Dict[str, dispatch] = defaultdict(lambda: [])



"""Module implementing core rules in Julia's dispatch system. For reference,
check Jeff Bezanson's PhD thesis at
https://github.com/JeffBezanson/phdthesis/blob/master/main.pdf
"""

import inspect
import itertools
import typing
from typing import Callable, Iterable, Optional, Tuple, Type, TypeVar, Union

from .poset import VT, PODict, POSet

T = TypeVar("T")


class Vararg:
    pass


class Sig:
    """Represent a Tuple type with extra features, mostly used for method signature
    dispatching"""

    def __init__(self, *types) -> None:
        self.types: Type[Tuple] = Tuple.__getitem__(types)
        self.arg_names: Tuple[str, ...] = tuple(["unset"] * len(types))
        self.func_site: Optional[str] = None

        arity = len(types)
        self.is_variadic: bool = arity and types[-1] is Vararg
        self.min_arity: int = arity - 1 if self.is_variadic else arity

    @property
    def __args__(self) -> Tuple[Type]:
        return self.types.__args__

    def __repr__(self) -> str:
        def name(t):
            origin = getattr(t, "__origin__", None)
            if origin is typing.Union:
                _names = " | ".join(x.__name__ for x in t.__args__)
                return f"[{_names}]"
            elif origin is type:
                return f"Type[{t.__args__[0].__name__}]"
            elif origin:
                return origin.__name__
            elif not hasattr(t, "__name__"):
                return repr(t)
            else:
                return t.__name__

        names = ", ".join(f":{name(t)}" for t in self.types.__args__)
        return f"Sig[{names}]"

    def __hash__(self) -> int:
        return hash(self.types)

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Sig):
            return self.types == o.types
        else:
            return self.types == o

    def dump(self) -> str:
        msg = repr(self) + f" at {self.func_site}"
        return msg


def get_origin(t) -> Type:
    if isinstance(t, Sig):
        return tuple

    if getattr(t, "__origin__", None):
        return t.__origin__


def is_variadic(t):
    try:
        return t.is_variadic
    except AttributeError:
        args = getattr(t, "__args__", tuple())
        last = args and t.__args__[-1] or None
        return last and (last is Vararg or getattr(last, "__origin__", None) is Vararg)


def min_arity(t, args, is_variadic):
    try:
        return t.min_arity
    except AttributeError:
        return len(args) - 1 if is_variadic else len(args)


def pad_varargs(a, b) -> Tuple[Tuple[Type], Tuple[Type]]:
    """Extract Tuple args and pad with `object`, accounting for varargs"""

    a_args = a.__args__
    a_var = is_variadic(a)

    b_args = b.__args__
    b_var = is_variadic(b)

    if not (a_var or b_var):
        return a_args, b_args

    na = min_arity(a, a_args, a_var)
    nb = min_arity(b, b_args, b_var)
    nm = max(na, nb)
    a_norm = tuple([*a_args[:na], *([object] * (nm - na))])
    b_norm = tuple([*b_args[:nb], *([object] * (nm - nb))])
    return a_norm, b_norm


def as_typetuple(t):
    return t.types if isinstance(t, Sig) else t


def issubtype(_type: Union[Type, Sig], _super: Union[Type, Sig]):
    """Check if `_type` is a subtype of `_super`.

    Arguments are either `Type` (including parameterized Type) or `Sig` (signature).

    Follow rules in 4.2.2 in Bezanson's thesis.

    Generic (aka parametric) types are invariant. For instance:
        `issubtype(List[int], List[object]) == False`
        `issubtype(List[int], List[int]) == True`
        `issubtype(list, List[object]) == False`

    The exceptions are:
        * `Tuple`: these are covariant. Eg:
            - `issubtype(Tuple[Foo], Tuple[Super]) == True` where `Foo -> Super`

        * `Union`: match if there's a covariant intersection, including non-union types
            - `issubtype(Foo, Union[str, Super]) == True` where `Foo -> Super`
            - `issubtype(str, Union[str, Super]) == True`

        * `typing.Type`: is covariant on the type argument. `type` is treated as
          `typing.Type[object]`
            - `issubtype(typing.Type[str], typing.Type[object]) == True`
            - `issubtype(typing.Type[str], type) == True`

    Since Python don't tag container instances with the type paramemeters
    (eg. `type([1,2,3]) == list`) this means that we can't dispatch lists as we would
    with arrays in Julia. The multiple dispatch system must then erase method signature
    generic information for such container types.

    See the `parametric` module for a way to declare and dispatch on parametric types.

    Also, we don't currently support the equivalent of `UnionAll`. This is not an issue
    since we don't support parametric dispatch, ie. `TypeVar`s in method signatures.
    """

    # quick test of equality
    if _type is _super:
        return True

    if _type == _super:
        return True

    # quick test of subclass
    try:
        if issubclass(_type, _super):
            return True
    except TypeError:
        pass

    # normalize Type[type]
    _type = typing.Type[object] if _type is type else _type
    _super = typing.Type[object] if _super is type else _super

    type_orig = get_origin(_type)
    super_orig = get_origin(_super)

    if (
        type_orig == super_orig
        and type_orig in (tuple, type)
        and super_orig in (tuple, type)
    ):
        # normalize params accounting for varargs
        type_params, super_params = pad_varargs(_type, _super)
        # check component-wise, Tuple types are covariant
        return len(type_params) == len(super_params) and all(
            issubtype(t, s) for (t, s) in zip(type_params, super_params)
        )

    elif type_orig is Union and super_orig is Union and _type == _super:
        # quick check for union equality
        return True

    elif type_orig is Union or super_orig is Union:
        # check for subtypes the rule is return true if there's a subtype relation
        # in the product of the parameters
        type_params = _type.__args__ if type_orig is Union else [_type]
        super_params = _super.__args__ if super_orig is Union else [_super]
        for (a, b) in itertools.product(type_params, super_params):
            if a is b or issubtype(a, b):
                return True
        return False

    return False


def is_more_specific(a, b):
    a_sub_b = issubtype(a, b)
    b_sub_a = issubtype(b, a)

    if a_sub_b and b_sub_a:
        if is_variadic(b):
            return True

    if b_sub_a:
        return False

    if a_sub_b:
        return True

    tag_a = getattr(a, "__tag__", None)
    tag_b = getattr(b, "__tag__", None)
    if tag_a is not None and tag_b is not None:
        return issubclass(tag_a, tag_b) and not issubclass(tag_b, tag_a)

    return False


def signatures_from(func: Callable) -> Iterable[Sig]:
    """Parse a callable to extract the dispatchable type tuple."""

    hints = typing.get_type_hints(func)
    kinds = {
        inspect.Parameter.POSITIONAL_ONLY,
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
        inspect.Parameter.VAR_POSITIONAL,
    }

    def get_type(p: inspect.Parameter):
        if p.kind == inspect.Parameter.VAR_POSITIONAL:
            return Vararg
        _type = hints.get(p.name, object)
        origin = getattr(_type, "__origin__", None)
        if origin and origin not in (typing.Union, type):
            return origin
        return _type

    params = list(
        p for p in inspect.signature(func).parameters.values() if p.kind in kinds
    )
    stack = [params]
    while stack:
        _types = []
        _names = []
        params = stack.pop()
        for p in params:
            _types.append(get_type(p))
            _names.append(p.name)

        sig = Sig(*_types)
        sig.arg_names = tuple(_names)

        code = func.__code__
        sig.func_site = f"{code.co_filename}:{code.co_firstlineno}"

        yield sig

        if params and params[-1].default != inspect.Parameter.empty:
            stack.append(params[:-1])


class SigDict(PODict[Sig, VT]):
    def __init__(self, data=None) -> None:
        super().__init__(data=data, lt=is_more_specific)


class SigSet(POSet[Sig]):
    def __init__(self, data) -> None:
        super().__init__(data=data, lt=is_more_specific)



import collections
import re
from contextlib import contextmanager
from typing import Any, Dict, Iterable, Optional

from gamma.config.load import load_node
from gamma.dispatch import dispatch
from ruamel.yaml.nodes import MappingNode, Node, SequenceNode

from . import tags
from .merge import merge_nodes
from .rawnodes import get_entry, get_id, get_keys, get_values
from .tags import Tag

SAFE_ENTRY_KEY = re.compile("^[A-Za-z0-9].+$")


class ConfigNode(collections.abc.Mapping):
    """Represent a dict-like config object.

    `ConfigNode` is immutable, changes should be made to the source `RootConfig`
    object. This class should be safe to pickle and pass to subprocesses.

    You when accessing keys by attribute `eg: config.foo` they'll return an empty
    `ConfigNode` instead of raising an `AttributeError`.


    """

    def __init__(
        self,
        node: MappingNode,
        root: Optional["RootConfig"] = None,
        parent: Optional["ConfigNode"] = None,
        key=None,
    ) -> None:
        """
        Args:
            node: the backing `MappingNode`
            root: the root of the config tree
            key: the node key in the config tree
        """

        self._node = node
        self._root = root
        self._key = key
        self._frozen = True
        self._parent = parent

    def __getitem__(self, key):
        ctx = dict(config=self, dump=False)
        return config_getitem(self, key, **ctx)

    def __iter__(self):  # pragma: no cover
        from .render import render_node

        return (render_node(key) for key in get_keys(self))

    def __len__(self) -> int:
        return config_len(self)

    def __getattr__(self, key):
        if key.startswith("__"):
            return object.__getattribute__(self, key)

        try:
            return self[key]
        except KeyError as err:
            raise AttributeError(key) from err

    def __call__(self, *args, **kwds):
        raise TypeError(
            "A ConfigNode object is not callable. Are you're trying to "
            f"access the empty/undefined key '{self._key}'?"
        )

    def __setattr__(self, name: str, value: Any) -> None:
        if not name[0] == "_":
            raise TypeError(
                f"'{self.__class__.__name__}' object does not support item assignment"
            )
        else:
            object.__setattr__(self, name, value)


class RootConfig(ConfigNode):
    """A root config object.

    The object is a collection of `(entry_key: str, entry: Node)`, sorted by
    the `entry_key`. A item access will search for the key in each entry and merge
    those found.

    New entries should be inserted with the push_entry` function.
    The entries are always iterated in `entry_key` lexicographical
    sort order and this affects merge results.
    """

    def __init__(self, entry_key: Optional[str] = None, entry=None) -> None:
        """
        Initialize the object, optionally adding a single entry. See
        [`push_entry`](api?id=push_entry).
        """
        self._root_nodes: Dict[str, MappingNode] = collections.OrderedDict()
        super().__init__(node=None, root=self, parent=None)

        if entry_key is not None:
            if entry is None:
                raise ValueError("Missing 'entry' argument")
            push_entry(self, entry_key, entry)


@dispatch
def push_entry(
    root: RootConfig, entry_key: str, entry, *, _allow_unsafe=False, content_type=None
) -> None:
    """Add an entry to the root config.

    The entry itself can be of any supported format by ``load_node``. You can
    disambiguate between string/bytes values by providing an optional ``content_type``
    """
    node = load_node(entry, content_type)
    push_entry(root, entry_key, node, _allow_unsafe=_allow_unsafe)


@dispatch
def push_entry(
    root: RootConfig, entry_key: str, node: Node, *, _allow_unsafe=False
) -> None:
    """Add a `Node` entry to the root config object.

    Entries are loaded in sorted order.

    Note: the global root object can only be modified by the thread that created it.
    """

    from .globalconfig import check_can_modify

    check_can_modify(root)

    if (not _allow_unsafe) and (not SAFE_ENTRY_KEY.match(entry_key)):
        pat = SAFE_ENTRY_KEY.pattern
        raise ValueError(f"Invalid entry_key: '{entry_key}'. Should match {pat}.")

    if entry_key in root._root_nodes:
        raise ValueError(f"Config file/entry named {entry_key} duplicated.")

    d = root._root_nodes
    d[entry_key] = node

    # sort by entry_key
    s = collections.OrderedDict(sorted(d.items(), key=lambda x: x[0]))
    root._root_nodes = s


@dispatch
def remove_entry(cfg: RootConfig, entry_key: str):
    """Remove an entry from the RootConfig object."""
    del cfg._root_nodes[entry_key]


@dispatch
def config_getitem(cfg: ConfigNode, key, **ctx):
    """Get an item from config by key."""
    _key, _item = get_entry(cfg._node, key)
    ctx = ctx.copy()
    ctx["key"] = _key
    return resolve_item(_item, **ctx)


@dispatch
def config_getitem(cfg: RootConfig, key, **ctx):
    """Get an item from a root config by key.

    We find all entries matching the key and merge them dynamically using
    `merge_nodes`.
    """
    matches = []
    key: str
    node: Node

    for entry_key, node in cfg._root_nodes.items():
        subkey, subnode = get_entry(node, key, default=None)
        if subnode:
            # append entry to the node comment
            matches.append((subkey, subnode))

    if matches:
        key, node = merge_nodes(matches)
    else:
        raise KeyError(key)

    return resolve_item(node, key=key, **ctx)


@dispatch
def resolve_item(item: Node, **ctx):
    """Resolve a config item from a ruamel.yaml `Node`

    This method delegates to a more specific method dispatched on (Node, Tag) types
    """
    tag = Tag[item.tag]()
    return resolve_item(item, tag, **ctx)


@dispatch
def resolve_item(item: Node, tag: Tag, **ctx):
    """Resolve a config item from a ruamel.yaml `Node`

    Fallback to rendering the node using `render_node`
    """

    from .render import render_node

    return render_node(item, tag, **ctx)


@dispatch
def resolve_item(item: MappingNode, tag: tags.Map, **ctx):
    """Wrap a plain `map` node as a child `ConfigNode` object"""
    cfg = ctx.get("config")
    root = cfg._root if cfg is not None else None
    return ConfigNode(item, root=root, parent=cfg)


@dispatch
def resolve_item(item: SequenceNode, tag: tags.Seq, **ctx):
    """Iterates on a `seq` node, resolving each child item node."""
    out = []
    for sub in get_values(item):
        out.append(resolve_item(sub, **ctx))
    return out


@dispatch
def get_keys(cfg: ConfigNode) -> Iterable[Node]:
    """Return all keys in a config node"""
    return get_keys(cfg._node)


@dispatch
def get_keys(cfg: RootConfig) -> Iterable[Node]:
    """Return all *distinct* keys in a config node"""
    returned = set()
    for node in cfg._root_nodes.values():
        for key in get_keys(node):
            key_id = get_id(key)
            if key_id in returned:
                continue
            returned.add(key_id)
            yield key


@dispatch
def config_len(cfg: ConfigNode) -> int:
    """Number of keys in a config node"""
    return len(cfg._node.value)


@dispatch
def config_len(cfg: RootConfig) -> int:
    """Number of *distinct* keys in a config node"""
    return len(list(get_keys(cfg)))


@dispatch
def create_last_entry_key(cfg: RootConfig) -> str:
    """Create an entry_key guaranteed to be the last entry for the object."""
    clen = config_len(cfg)
    return f"~{clen:03d}-temp"


@dispatch
@contextmanager
def config_context(cfg: RootConfig, partial) -> None:
    entry_key = create_last_entry_key(cfg)
    push_entry(cfg, entry_key, partial, _allow_unsafe=True)
    try:
        yield
    finally:
        remove_entry(cfg, entry_key)


@dispatch
@contextmanager
def config_context(partial) -> None:
    from gamma.config.globalconfig import get_config

    cfg = get_config()
    with config_context(cfg, partial):
        yield



"""Implements the node merging functionality"""
import re
from copy import deepcopy
from functools import reduce
from typing import List, Tuple, Union

from gamma.dispatch import dispatch
from ruamel.yaml.nodes import MappingNode, Node, SequenceNode

from .rawnodes import get_item, get_keys, get_values, is_in, union_nodes

hints_pattern = re.compile("^.*@hint: *?([A-Za-z0-9_]+) ?.*$")


@dispatch
def merge_nodes(nodes: List):
    """Merge nodes iterable, ignoring key"""
    if len(nodes) == 0:
        raise ValueError("Empty nodes list")
    elif len(nodes) == 1 and isinstance(nodes[0], Tuple):
        return nodes[0]
    elif len(nodes) == 1 and isinstance(nodes[0], Node):
        return None, nodes[0]
    return reduce(merge_nodes, nodes)


@dispatch
def merge_nodes(l_node: MappingNode, r_node: MappingNode):
    """Merge map nodes ignoring key"""
    return merge_nodes(None, l_node, None, r_node)


@dispatch
def merge_nodes(left: Tuple, r_node: MappingNode):
    """Merge map nodes ignoring key (for assymetric fold-left)"""
    l_node = left[-1]
    return merge_nodes(None, l_node, None, r_node)


@dispatch
def merge_nodes(left: Tuple, right: Tuple):
    """Merge map nodes (symetric fold-left)

    Args:
        left, right: Tuple of (key: Node, value: Node)
    """

    l_key, l_node = left
    r_key, r_node = right
    return merge_nodes(l_key, l_node, r_key, r_node)


@dispatch
def merge_nodes(l_key, l_node: Union[None, Node], r_key, r_node: Union[None, Node]):
    """Merge scalars.

    Return "right", or "left" if "right" is `None`
    """
    # scalars,
    if r_node is not None:
        return r_key, r_node
    else:
        return l_key, l_node


@dispatch
def merge_nodes(l_key, l_node: MappingNode, r_key, r_node: MappingNode):
    """Merge `map` nodes recursively.

    Right side has precedence. `@hint: merge_replace` overrides merging, returning
    right.
    """
    if has_replace_hint(r_node):
        return r_key, r_node

    newvalue = []
    # Merge if both nodes are mappings
    l_keys = get_keys(l_node)
    r_keys = get_keys(r_node)
    subkeys = union_nodes(r_keys, l_keys)
    for subkey in subkeys:
        l_subnode = get_item(l_node, subkey, default=None)
        r_subnode = get_item(r_node, subkey, default=None)
        _, subvalue = merge_nodes(subkey, l_subnode, subkey, r_subnode)
        newvalue.append((subkey, subvalue))

    # by default, we use right side meta data
    newnode = deepcopy(r_node)
    newnode.value = newvalue
    return r_key, newnode


@dispatch
def merge_nodes(l_key, l_node: SequenceNode, r_key, r_node: SequenceNode):
    """Merge `sequence` nodes recursively.

    Right side has precedence. `@hint: merge_replace` overrides merging, returning
    right.
    """

    if has_replace_hint(r_node):
        return r_key, r_node

    newvalue = list(get_values(l_node)).copy()
    for r_item in get_values(r_node):
        if not is_in(r_item, newvalue):
            newvalue.append(r_item)

    # by default, we use right side meta data
    newnode = deepcopy(r_node)
    newnode.value = newvalue
    return r_key, newnode


@dispatch
def has_replace_hint(node: Node) -> bool:
    """Check for existence of config "hint" in the node's comments.

    Only `@hint: merge_replace` is supported.
    """

    node_comments = getattr(node, "comment", [])
    if not node_comments:
        return False

    # flatten comment hierarchy. can't use itertools.chain here :(
    hints = set()
    stack = [node_comments]
    while stack:
        item = stack.pop(0)

        if item and hasattr(item, "value"):
            match = hints_pattern.match(item.value)
            if match:
                hints.add(match.group(1))

        elif isinstance(item, list):
            for sub in item:
                if sub:
                    stack.append(sub)

    return "merge_replace" in hints



# flake8: noqa
from .dispatchsystem import DispatchError, dispatch
from .parametric_types import ParametricMeta, Val, parametric


def _isjupyter():  # pragma: no cover
    try:
        shell = get_ipython().__class__.__name__  # type: ignore
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        return False
    except NameError:
        return False  # Probably standard Python interpreter


def install_breakpoint_hook():

    if _isjupyter():
        # we don't like Jupyter
        return

    import sys
    import pdb as orig_pdb

    pdb = orig_pdb.Pdb(
        skip=["gamma.dispatch.*", "typing", "abc", "inspect", "bdb", "pdb"]
    )
    sys.breakpointhook = pdb.set_trace



from typing import Mapping, Optional, Tuple

from pydantic import BaseModel

NO_DISCRIMINATOR = object()


class ConfigStruct(BaseModel):
    __discriminator__ = "kind"

    @classmethod
    def get_subclasses(cls):
        yield cls
        stack = [cls]
        while stack:
            cc = stack.pop(0)
            for sub in cc.__subclasses__():
                stack.append(sub)
                yield sub

    @classmethod
    def parse_scalar(cls, value):
        kind = value
        args = {cls.__discriminator__: value}
        return kind, args

    @classmethod
    def get_discr_value(cls):
        key = cls.__discriminator__
        try:
            value = getattr(cls, key)
            return value
        except AttributeError:
            pass

        try:
            _field = cls.__fields__[key]
            value = getattr(_field, "default", None)
            if value is None:
                raise ValueError(f"Field {cls.__name__}.{key} has no value set")
            return value
        except KeyError:
            pass

        return NO_DISCRIMINATOR

    @classmethod
    def parse_obj(cls, obj):
        if not isinstance(obj, Mapping):
            ret = cls.parse_scalar(obj)
            if isinstance(ret, Tuple):
                discr, args = ret
            else:
                return ret
        else:
            try:
                discr = obj[cls.__discriminator__]
                args = obj
            except KeyError:
                raise ValueError(f"Missing '{cls.__discriminator__}' field.")

        _type = cls.find_type(discr)
        return _type(**args)

    @classmethod
    def find_type(cls, discr):
        match = [k for k in cls.get_subclasses() if k.get_discr_value() == discr]
        if not match:
            raise ValueError(
                f"No {cls.__name__} for {cls.__discriminator__} = '{discr}'"
            )
        return match[0]

    @classmethod
    def __get_validators__(cls):
        yield cls.parse_obj

    class Config:
        frozen = True


class URIConfigStruct(ConfigStruct):
    __discriminator__ = "scheme"
    scheme: str
    uri: Optional[str]

    @classmethod
    def parse_obj(cls, obj):
        if isinstance(obj, Mapping):
            if "uri" in obj and cls.__discriminator__ not in obj:
                scheme, _ = cls.split_uri(obj["uri"])
                obj["scheme"] = scheme

            if "uri" in obj and hasattr(obj, "update"):
                obj2 = cls.parse_uri(obj["uri"]).dict()
                obj.update(obj2)

        return super().parse_obj(obj)

    @staticmethod
    def split_uri(value):
        value = str(value)
        if ":" not in value:
            raise ValueError(f"'{value}' not a valid URI: missing ':'")
        return value.split(":", 1)

    @classmethod
    def parse_scalar(cls, value):
        scheme, _ = cls.split_uri(value)
        _type: URIConfigStruct = cls.find_type(scheme)
        return _type.parse_uri(value)

    @classmethod
    def parse_uri(cls, uri: str):
        return cls(uri=uri)



"""Base render_node methods"""

import logging
from typing import Any, Optional

from gamma.config.confignode import ConfigNode, RootConfig  # noqa
from gamma.dispatch import dispatch
from ruamel.yaml.nodes import MappingNode, Node, ScalarNode, SequenceNode

from . import tags
from .rawnodes import get_entries, get_values

logger = logging.getLogger(__name__)


class RenderDispatchError(Exception):
    pass


@dispatch
def render_node(
    node: Node,
    tag: tags.Tag,
    *,
    key: Optional[Node] = None,
    config: "ConfigNode" = None,
    dump: bool = False,
    recursive: bool = False,
    path: Optional[str] = None,
):
    """Spec for tag handling functions.

    This function is dispatched first on the specific (Node, Tag) types. If the tag
    name contains a `:` (colon) character, the tag name is interpreted as a URI and
    we fallback to dispatching by the "scheme" portion of the URI and add an extra
    `path` kwarg to the method call. Example:

    ```
    foo: !mytag 1
    ```

    Will be dispatched solely on (ScalarNode, Tag["!mytag"]) types. Whereas

    ```
    foo: !mytag:bar 1
    ```

    Will first be dispatched on (ScalarNode, Tag["!mytag:bar"]), then on
    (ScalarNode, Tag["!mytag"]), the last one with the optional `path` kwargs filled
    as `bar`

    Args:
        node: The raw ruamel YAML node. Can be returned during a dump to keep
            the field dynamic or hide sensitive values.
        tag: The tag to be handled. A subinstance of ``Tag``.
    Keywork Args:
        key: The node key, also as a ruamel YAML node.
        config: The current ConfigNode object.
        dump: Flag indicating we're dumping the data to a potentially insecure
            destination, so sensitive data should not be returned.
        recursive: If true, tag handlers should recursively render nodes.
        path: The URI path for URI fallback dispatch

    Return:
        any value
    """

    if ":" in tag.name:
        # try scheme based dispatch
        scheme, path = tag.name.split(":", 1)
        SchemeTag = tags.Tag[scheme]
        return render_node(
            node, SchemeTag(), key=key, config=config, dump=dump, path=path
        )

    # not found errror
    msg = f"""Renderer not implemented for node_type={type(node).__name__}, tag={tag.name}
    Possible causes:
        * Check if you forgot to import `gamma.config.render_node` when defining your
          `def render_node` method.
    """  # noqa
    raise RenderDispatchError(msg)


@dispatch
def render_node(node: Node, **args) -> Any:
    """Render node.

    Construct a parameterized `Tag` class from `node.tag` and call
    `render_node(Node, Tag)`
    """

    TagClass = tags.Tag[node.tag]
    return render_node(node, TagClass(), **args)


@dispatch
def render_node(node: ScalarNode, tag: tags.Str, **args):
    """Render scalar string"""
    return node.value


@dispatch
def render_node(node: ScalarNode, tag: tags.Int, **args):
    """Render scalar int"""
    return int(node.value)


@dispatch
def render_node(node: ScalarNode, tag: tags.Float, **args):
    """Render scalar float"""
    return float(node.value)


@dispatch
def render_node(node: ScalarNode, tag: tags.Null, **args):
    """Render scalar null"""
    return None


@dispatch
def render_node(node: ScalarNode, tag: tags.Bool, **args):
    """Render scalar boolean. Accepts YAML extended interpretation of `bool`s, like
    `ruamel.yaml`
    """

    val = node.value.lower()
    if val in ("y", "yes", "true", "on", "1"):
        return True
    elif val in ("n", "no", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"Can't convert '{node.value}' to bool")


@dispatch
def render_node(node: ScalarNode, tag: tags.Timestamp, **args):
    """Render timestamp node as `datetime.datetime`. Use `dateutil.parser` module"""
    from dateutil import parser

    return parser.parse(node.value)


@dispatch
def render_node(node: SequenceNode, tag: tags.Seq, **args):
    """Render `seq` nodes recursively"""
    subargs = args.copy()

    out = []
    for subvaluenode in get_values(node):
        subvalue = render_node(subvaluenode, **subargs)
        out.append(subvalue)

    return out


@dispatch
def render_node(node: MappingNode, tag: tags.Map, **args):
    """Render `map` nodes recursively"""
    subargs = args.copy()

    out = {}
    for subkeynode, subvaluenode in get_entries(node):
        subkey = render_node(subkeynode, **args)
        subargs["key"] = subkeynode
        subvalue = render_node(subvaluenode, **subargs)
        out[subkey] = subvalue

    return out


@dispatch
def render_node(cfg: "RootConfig", **args):
    """Render the resulting node of merging all entries"""

    from gamma.config.merge import merge_nodes

    nodes = list(cfg._root_nodes.values())
    _, node = merge_nodes(nodes)
    args.setdefault("config", cfg)
    args.setdefault("dump", False)
    return render_node(node, **args)


@dispatch
def render_node(cfg: "ConfigNode", **args):
    """Render the config node.

    Args:
        dump: If true, assume it's "dump mode" where secrets are not to be rendered.
    """
    args.setdefault("config", cfg)
    args.setdefault("dump", False)
    return render_node(cfg._node, **args)


from . import builtin_tags  # noqa isort:skip
